
/* *************************************************************************
 *  File : cyba.c
 * -------------------------------------------------------------------------
 *  All rights are reserved by Bo-kyun Na, IntellActive Research Inc.
 * -------------------------------------------------------------------------
 * This example reads an image from /dev/video, transmits compressed video 
 * over the local network, optionally displays the image in Xwindow, and 
 * also, receives, uncompresses, then displays images.
 * ************************************************************************* */

				// layout.moc should be before X11 headers
#include "layout.moc"		// gui: from layout.h by Qt 

#include <semaphore.h>
#include <iostream.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "ipc.h"		// inter-process communication headers
//#include "shmemblock.h"		// shared memory variables
#include "params.h"		// required parameters
#include "codec.h"		// codec macros
#include "network.h"		// network headers
#include "transceive.h"		// sned, receive classes
#include "pear.h"          	// pthread macroing 
#include "vdo.h"		/* video related header */
#include "odo.h"		/* audio related header */


/* global variables */
int verbose = 0;
sem_t VS, VR, AS, AR;

/* function prototypes */
void error ();

void TELL_WAIT ();		// signal to wait in multiprocessing
void TELL_PARENT ();		// signal to parent that I am done.
void WAIT_PARENT ();		// signal waiting when parent has done.
void TELL_CHILD (pid_t pid);	// siganl to child parent done.
void WAIT_CHILD ();		// signal waiting when child has done.


/* *************************************************************************
 * Function    : 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */

/* *************************************************************************
 * usage ()
 *
 *
 * ************************************************************************* */
void usage (char *pgm)
{
  fprintf (stderr, "usage: %s [options] \n \
Legal options:                                       Defaults\n \
<-i file_name> : audio input file name               0\n \
<-o file_name> : audio output file name              0\n \
", pgm);

  exit (1);
}


/* *************************************************************************
 * Function    : void *Thread_TxVideo (void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *Thread_TxVideo (void *arg)
{
  //  Parameters *param = (Parameters *) shm->param;
  SharedVariables *shm = (SharedVariables *) arg;
  Parameters *param = new Parameters;
  param->hostname = shm->param.hostname;
  param->hostnameknt = shm->param.hostnameknt;
  param->channel = shm->param.channel;
  param->transmit = shm->param.transmit;

  param->dev_vdo = shm->param.dev_vdo;

  param->frames = shm->param.frames;
  param->width = shm->param.width; 
  param->height = shm->param.height; 
  param->depth = shm->param.depth;


  TransceiveState *tstate;

  int vid;
  unsigned char *capimg;	/* buffer for captured image */
  XWA xwa;
  V4L2 v4l2;
  VIMAGE vimage;

  int first = 0;
  int CISsize;
  unsigned char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 
  unsigned char outbuf[VDO_WIDTH * VDO_HEIGHT];
  unsigned char *y, *u, *v;

  /* Pointers for the conversion of RGBtoYUV */
  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, VIDEO);

  /* Init. of xwindow depth */
  xwa.xdepth = param->depth; 
  if (verbose)
    printf ("xdepth on TV is %d\n", xwa.xdepth);

  /* Function: Init. of video capture to get the video from my camera */
  vid = BV_init_vidcap (param->dev_vdo, &v4l2);

  /* Init. and format of xwindow display */
  vimage.width = VDO_WIDTH;
  vimage.height = VDO_HEIGHT;
  BV_init_display(&xwa, param->argc, param->argv, &vimage);
  if (vimage.data == NULL) {
    printf("malloc(%d) for vimage.data failed\n", v4l2.fmt.fmt.pix.sizeimage);
    exit (1);
  }

  capimg = (unsigned char *) malloc (v4l2.fmt.fmt.pix.sizeimage);
  if (capimg == NULL) {
    printf("malloc(%d) for capimg failed\n", v4l2.fmt.fmt.pix.sizeimage);
    exit (1);
  }

//    /* Initialization of codec functions */
//    if (param->codec == H261) {
//      init_codec ();		/* Initialization of codec of H261 */
//      init_encoder ();		/* Initilization of H261 encoder */
//    }
//    else if (param->codec == H263) {
//      //    H263initEncoder(VDO_WIDTH, VDO_HEIGHT);
//    }

  /* Loop processing related to a video frame */
  for (;;) {
    //    sem_wait (&VS);
    /* to get captured video images */
    BV_get_vdoframe (vid, (char *) capimg, &v4l2);

    /* Packetization and sending of video data */
    if (param->transmit) {
      /* Conversion image pixel format */
      RGB16toYUV(capimg, y, u, v, VDO_WIDTH, VDO_HEIGHT);

      if (param->codec == H261) {
	CISsize = ((first %= 10) == 0) ? 
	  p64EncodeIntraFrame ((char *) inbuf, (char *) outbuf):
	  p64EncodeFrame ((char *) inbuf, (char *) outbuf);

	/* transmit compressed image sequence data */
	BV_packetize_vdoframe(outbuf,CISsize,VDO_WIDTH,VDO_HEIGHT,tstate,param);
      }
      else if (param->codec == H263) {
	//	CISsize = ((first %= 10) == 0) ? H263codeIntraFrame(frame, buf) :
	//	                                 H263codeFrame(&frame, buf);
       //BV_packetize_vdoframe(outbuf,CISsize,VDO_WIDTH,VDO_HEIGHT,tstate,param);
      }
      else if (param->codec == JPEG) {
	printf ("Not available.\n");
	exit (0);
      }
      else if (param->codec == MPEG) {
	printf ("Not available.\n");
	exit (0);
      }
      else if (param->codec == RAW) {	/* transmit raw video data */
	BV_packetize_vdoframe (capimg, v4l2.fmt.fmt.pix.sizeimage, 
			       VDO_WIDTH, VDO_HEIGHT, tstate, param);
      }
      else {
	printf ("Error on sending. Specifiy a correct codec type.\n");
	exit (0);
      }
      first ++;   

      /* Color conversion to fit on your display */
      if ((xwa.xdepth == 24) || (xwa.xdepth == 32))
	RGB16toRGB32(capimg, vimage.data, VDO_WIDTH, VDO_HEIGHT);
      else if (xwa.xdepth == 16)
	memcpy (vimage.data, capimg, v4l2.fmt.fmt.pix.sizeimage);

      /* Display compressed local images */
      if (shm->lvdo_flag)
	BV_display_vdoframe (&xwa, &vimage);
    }

    /* Store compressed video data as a file */
    //    if (param->wrfile != NULL)	
    //      fwrite(vimage.data, 1, CISsize, fp);

    //    sem_post (&VR);
    //    thrd_stop (&AS);   /* Thread stop */
  }
}


/* *************************************************************************
 * Name        : void *Thread_RxVideo (void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *Thread_RxVideo (void *arg)
{
  int userno = 0;
  unsigned char *y, *u, *v;
  unsigned char inbuf[VDO_WIDTH * VDO_HEIGHT * 3/2]; 
  unsigned char outbuf[VDO_WIDTH * VDO_HEIGHT * 4]; 

  XWA xwa;
  VIMAGE vimage;

  Parameters *param = (Parameters *) arg;
  TransceiveState *receiver;
  MsgType *message;


  /* set pointers of codec buffer */
  y = inbuf;
  u = &inbuf[VDO_WIDTH * VDO_HEIGHT];
  v = &inbuf[VDO_WIDTH * VDO_HEIGHT * 5/4];

  /* Initialization of network environments */
  receiver = initialize_receive(param, VIDEO);

  message = (MsgType *) malloc (sizeof (MsgType));
  if (message == NULL) {
    printf ("malloc fail: message = %d\n", sizeof (MsgType));
    exit (1);
  }

  /* Init. of xwindow depth */
  xwa.xdepth = param->depth; 
  if (verbose)
    printf ("xdepth on RV is %d\n", xwa.xdepth);

  /* Initialization of image display  */
  vimage.width = VDO_WIDTH;
  vimage.height = VDO_HEIGHT;
  BV_init_display(&xwa, param->argc, param->argv, &vimage);

//    if (param->codec == H261) {
//      init_codec ();		/* Initialization of codec of H261 */
//      init_decoder ();		/* Initilization of H261 decoder */
//    }
//    else if (param->codec == H263) {
//      //    H263initDecoder();
//    }

  if (param->wrfile != NULL) 
    param->fp = fopen(param->wrfile, "w");

  for (;;) {
    //    sem_wait (&VR);

    message = (MsgType *) BV_VDO_receive (receiver, message);

    if (param->codec == H261) {
      /* Decode of recvd message and conversion to RGB from YUV  */
      decode_frame ((char *)message->data, message->header.size, (char *)inbuf,
      		    ((message->header.frame % 10) == 0) ? 1 : 0, userno);

      /* Color Conversion YUV411 to RGB32 or 16 (RGB565) */
      if ((xwa.xdepth == 24) || (xwa.xdepth == 32)) {
	YUV411toRGB32 (y, u, v, vimage.data, VDO_WIDTH, VDO_HEIGHT);

      }
      else if (xwa.xdepth == 16) { 
	YUV411toRGB32 (y, u, v, outbuf, VDO_WIDTH, VDO_HEIGHT);
	RGB32toRGB16 (outbuf, vimage.data, VDO_WIDTH, VDO_HEIGHT);
      }
    } 
    else if (param->codec == H263) {
      //      inbuf = 
     //(unsigned char *)H263decodeFrame(0, message->data, message->header.size);

      /* Color Conversion YUV411 to RGB32 or 16 (RGB565) */
      if ((xwa.xdepth == 24) || (xwa.xdepth == 32)) {
	YUV411toRGB32 (y, u, v, vimage.data, VDO_WIDTH, VDO_HEIGHT);
      }
      else if (xwa.xdepth == 16) { 
	YUV411toRGB32 (y, u, v, outbuf, VDO_WIDTH, VDO_HEIGHT);
	RGB32toRGB16 (outbuf, vimage.data, VDO_WIDTH, VDO_HEIGHT);
      }
    }
    else if (param->codec == JPEG) {
      printf ("Not yet served this codec\n");
      exit (0);
    }
    else if (param->codec == MPEG) {
      printf ("Not yet served this codec\n");
      exit (0);
    }
    else if (param->codec == RAW) {
      /* display local video image */
      if ((xwa.xdepth == 24) || (xwa.xdepth == 32))
	RGB16toRGB32(message->data, vimage.data, VDO_WIDTH, VDO_HEIGHT);
      else if (xwa.xdepth == 16)
	memcpy(vimage.data, message->data, message->header.size); 
    }
    else {
      printf ("Error on compression type: Specify the codec type\n");
      exit (0);
    }

    /* Display images */
    BV_display_vdoframe (&xwa, &vimage);

    /* Store received video data as a file */
    //      if (param->wrfile)	
    //	fwrite(vimage.data, 1, message.header.size, df);



    //    sem_post (&VS);
    //  thrd_stop (&AS); 
  }
}



/* *************************************************************************
 * Function    : void *Thread_TxAudio (void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *Thread_TxAudio (void *arg)
{
  int fd;			/* sound device file descriptor */
  //  int status;			/* return status of system calls */
  int rd;			/* size in byte unit */
  int odo_frame;

  unsigned char *buf;		/* this buffer holds the digitized audio */

  /* Network parameters */
  //  Parameters *param = (Parameters *) arg;
  TransceiveState *tstate;


  SharedVariables *shm = (SharedVariables *) arg;
  Parameters *param = new Parameters;
  param->hostname = shm->param.hostname;
  param->hostnameknt = shm->param.hostnameknt;
  param->channel = shm->param.channel;
  param->transmit = shm->param.transmit;
  param->dev_odo = shm->param.dev_odo;
  param->frames = shm->param.frames;
  param->width = shm->param.width; 
  param->height = shm->param.height; 
  param->depth = shm->param.depth;


  odo_frame = ODO_LENGTH * ODO_RATE * ODO_SIZE * ODO_CHANNELS / 8;
  buf = (unsigned char *) malloc(odo_frame);
  if (buf == NULL) {
    printf ("malloc fail: sizeof (buf) = %d\n", odo_frame);
    exit (1);
  }

  /* Initialization of transmission parameters */
  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, AUDIO);

  sem_wait (&AS);		

  /* Initialization of audio */
  fd = BV_ODO_initialize (param->dev_odo);

  for ( ; ; ) {			/* loop until Control-C */
    rd = BV_ODO_record (fd, buf, odo_frame);
    if (verbose) printf ("Recorded audio size is %d\n", rd);
    //    close (fd);

    BV_odo_packetize (buf, odo_frame, tstate, param);

    //    sem_post (&AR);		/* only for audio */
    //    thrd_stop (&VS);
  }
}



/* *************************************************************************
 * Function    : Thread : void *Thread_RxAudio(void *arg)
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void *Thread_RxAudio (void *arg)
{
  int fd;			/* sound device file descriptor */
  int wr;			/* played audio size in byte unit */
  int status;			/* return status of system calls */
  int odo_frame;		/* size of audio buffer */

  unsigned char *buf;		/* this buffer holds the digitized audio */

//    Parameters *param = (Parameters *) arg;
  TransceiveState *RxState;
  MsgType *message;

  SharedVariables *shm = (SharedVariables *) arg;
  Parameters *param = new Parameters;
  param->hostname = shm->param.hostname;
  param->hostnameknt = shm->param.hostnameknt;
  param->channel = shm->param.channel;
  param->transmit = shm->param.transmit;
  param->dev_odo = shm->param.dev_odo;
  param->frames = shm->param.frames;
  param->width = shm->param.width; 
  param->height = shm->param.height; 
  param->depth = shm->param.depth;


  odo_frame = ODO_LENGTH * ODO_RATE * ODO_SIZE * ODO_CHANNELS / 8;
  buf = (unsigned char *) malloc(odo_frame);
  if (buf == NULL) {
    printf ("malloc fail: sizeof (buf) = %d\n", odo_frame);
    exit (1);
  }

  message = (MsgType *) malloc (sizeof (MsgType));
  if (message == NULL) {
    printf ("malloc fail: message = %d\n", sizeof (MsgType));
    exit (1);
  }

  /* Init. of network */
  RxState = (TransceiveState *) initialize_receive (param, AUDIO);

  sem_wait (&AR);

  /* Initialization of audio */
  fd = BV_ODO_initialize (param->dev_odo);

  /* loop for audio play until Control-C */
  for ( ; ; ) {		
    message = (MsgType *) BV_ODO_receive (RxState, message);

    /* play audio data */
    wr = BV_ODO_play (fd, message->data, odo_frame);
    if (verbose) printf ("Played audio size is %d bytes\n", wr);

    /* wait for playback to complete before recording again */
    status = ioctl(fd, SOUND_PCM_SYNC, 0); 
    if (status == -1)
      perror("SOUND_PCM_SYNC ioctl failed");
    //    close (fd);

    //    sem_post (&AS);
    //    thrd_stop (&VS);
  }
}


/* *************************************************************************
 * Function    : int main (int argc, char **argv) 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
int main (int argc, char **argv)
{
  int   i;
  pid_t pid;

  Parameters *param;

  pthread_t tid_VidTx,
            tid_VidRx,
            tid_AudRx, 
            tid_AudTx;

  pthread_attr_t attr;

  //  SEM_INIT (&VS, 0, 0);
  //  SEM_INIT (&VR, 0, 0);
  SEM_INIT (&AS, 0, 0);
  SEM_INIT (&AR, 0, 0);

  PTHREAD_ATTR_INIT (&attr);
  pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);

  /* Initialization of video and audio parameters  */
  param = (Parameters *) init_Param(argc, argv);
  param->argc = argc;
  param->argv = argv;

  /* audio device node name */
  strcpy (param->dev_odo, ODO_DEVICE);		

  /* video device node name */
  strcpy(param->dev_vdo, VDO_DEVICE);

  /* Init. of video color processing */
  init_dither_tab ();

  /* Process command line arguments */
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-C")) {
      if (!strcmp(argv[++i], "H261")) {
      	param->codec = H261;
      	param->cis_type = "H261";
      }
      else if (!strcmp(argv[i], "H263")) {
      	param->codec = H263;
      	param->cis_type = "H263";
      }
      else if (!strcmp(argv[i], "JPEG")) {
      	param->codec = JPEG;
      	param->cis_type = "Jpeg";
      }
      else if (!strcmp(argv[i], "MPEG")) {
      	param->codec = MPEG;
      	param->cis_type = "Mpeg1";
      }
      else if (!strcmp(argv[i], "RAW")) {
	param->codec = RAW;
      	param->cis_type = "RAW";
      }
      else {
	printf ("Specifiy your codec mode\n");
        usage(argv[0]);
      }
    }
    else if(!strcmp(argv[i], "-D")) {
      strcpy (param->dev_vdo, argv[++i]);
    }
    else if(!strcmp(argv[i], "-x")) {
      param->depth = atoi(argv[++i]);
      if (!((param->depth == 15) || (param->depth == 16) || (param->depth == 24)
	  || (param->depth == 32))) {
	printf("Unrecognized display depth. Type correct depth\n");
        usage(argv[0]);
      }
    }
    else if (!strcmp(argv[i], "-i")) {
      strcpy (param->rdfile, argv[++i]);
      param->fp = fopen(param->rdfile, "r");
      if (param->fp == NULL) 
	printf ("Can't open image file: %s\n", param->rdfile);
      else  printf ("Image file, \"%s\", is opened.\n", param->rdfile);
    }
    else if (!strcmp(argv[i], "-o")) {
      strcpy (param->wrfile, argv[++i]);
      param->fp = fopen(param->wrfile, "w");
      if (param->fp == NULL) 
	printf ("Can't open image file: %s\n", param->wrfile);
      else  printf ("Image file, \"%s\", is opened\n", param->wrfile);
    }
    // else if(!strcmp(argv[i], "-r")) {
    //    rate = atoi(argv[++i]);
    //    }
    else if(!strcmp(argv[i], "-s")) {
      param->width = atoi(argv[++i]);
      param->height = atoi(argv[++i]);
    }
    else if(!strcmp(argv[i], "-v")) {
      verbose = 1;
    }
  }

  /* Initialization of codec functions */
  if (param->codec == H261) {
    init_codec ();		/* Initialization of codec of H261 */
    init_encoder ();		/* Initilization of H261 encoder */
    init_decoder ();		/* Initilization of H261 decoder */
  }
  else if (param->codec == H263) {
    //    H263initEncoder(VDO_WIDTH, VDO_HEIGHT);
    //    H263initDecoder();
  }

  TELL_WAIT ();

  //  sem_post (&shm->AR);

  if ((pid = fork ()) < 0) {
    cout << "fork error" << endl;
    exit (1);
  }
  else if (pid > 0) {		/* parent process */
    // allocate a shared memory block
    ipcSharedMemory mb( ".BVrc", sizeof(SharedVariables) );
    if (!mb) {
      cerr << "error # " << mb.rdstate()
	   << " allocate a shared memory block"
	   << endl;
      return 1;
    }

    // get pointer to block and fill in structure
    SharedVariables *p = (SharedVariables *) mb.Pointer();
    if (!p) {
      cerr << "could not get pointer" << endl;
      return 2;
    }

    // Initialize shared variables into the shared memory
    //    memcpy(p->param, param, sizeof (Parameters));

    TELL_CHILD (pid);		

    /* Creation of receive video threads */
    sleep (1);
    PTHREAD_CREATE (&tid_VidRx, &attr, Thread_RxVideo, (void *) param);

    sleep (1);
    PTHREAD_CREATE (&tid_AudRx, &attr, Thread_RxAudio, (void *) param);  
  }
  else {			/* child process */
    WAIT_PARENT ();		// wait until parent initialized

    ipcSharedMemory mb(".BVrc"); // access existing shared memory block
    if (!mb) {
      cerr << "error # " << mb.rdstate() 
	   << " access existing shared memory block" 
	   << endl;
      return 1;
    }

    // get pointer to block and print values
    SharedVariables *shm = (SharedVariables *) mb.Pointer();
    if (!shm) {
      cerr << "could not get pointer" << endl;
      return 2;
    }

    shm->param.hostname = param->hostname;
    shm->param.hostnameknt = param->hostnameknt;
    shm->param.channel = param->channel;
    shm->param.transmit = param->transmit;

    shm->param.dev_vdo = param->dev_vdo;
    shm->param.dev_odo = param->dev_odo;

    shm->param.frames = param->frames;
    shm->param.width = param->width; 
    shm->param.height = param->height; 
    shm->param.depth = param->depth;

    /* Creation of send/receive video threads */
    sleep (1);
    PTHREAD_CREATE (&tid_VidTx, &attr, Thread_TxVideo, (void *) shm);

    sleep (1);
    //    PTHREAD_CREATE (&tid_AudTx, &attr, Thread_TxAudio, (void *) shm);

    // Graphical User Interface
    QApplication a ( argc, argv );
    BVWidget *f;
    f = new BVWidget (shm);
    f->setCaption ("BuenaVista");
    a.setMainWidget (f);
    f->show ();
    a.connect ( &a, SIGNAL(lastWindowClosed()), &a, SLOT(quit()) );
    a.exec();

  }

  pthread_exit ((void *) NULL);
}



